/*
Questo sketch disegna un robot mobile di tipo 'uniciclo'
e lo controlla in velocità (senza limiti di saturazione)
verso il punto cliccato col mouse: implementa la legge
di controllo proporzionale.

Sulla schermata vengono riportati: le coordinate
(x,y,theta) dell'uniciclo, le coordinate (xDes,yDes)
del punto da raggiungere e il tempo impiegato per la
missione. Inoltre si riportano le velocità angolari 
(omegaR, omegaL) delle due ruote, le loro accelerazioni 
(omegaRp,omegaLp), la velocità longitudinale v1 e quella
angolare v2.
*/

// Coordinate attuali uniciclo
float x = 0;
float y = 0;
float theta = 0;

// Coordinate desiderate uniciclo
float xDes = 0;
float yDes = 0;

// Caratteristiche fisiche uniciclo
float r = 8; // raggio ruote attuate in pixel
float d = 25; // distanza tra le ruote attuate in pixel
float W = 5; // spessore ruote attuate in pixel
float R = 1.2*sqrt(pow(r,2)+pow(d/2+W/2,2)); // raggio robot

float dt = (float) 1/60; // tempo di campionamento

float e_p = 0; // errore posizionamento
float v1 = 0; // velocità lineare robot
float kv1 = 2; // costante legge proporzionale controllo v1
float v2 = 0; // velocità rotazionale robot
float kv2 = 5;  // costante legge proporzionale controllo v2
int nGiri = 0; // conta quanti giri su se stesso ha fatto il robot
float thetaDes; // orientamento desiderato (per raggiungere il target)
float t0,tempo; // tempo inizio e attuale missione

float omegaR = 0; // velocità angolare ruota destra
float omegaL = 0; // velocità angolare ruota sinistra
float omegaRp, omegaLp; // accelerazioni angolari ruote

void setup() 
{
  size(1250,700);
}

void draw() 
{
  
  background(0);

  pushMatrix();
  translate(width/2,height/2); // metto l'origine al centro della finestra

  if (mousePressed) // assegno target
  {
    xDes = mouseX - width/2;
    yDes = height/2 - mouseY;
    t0 = millis(); // inizio conteggio del tempo di missione
  }

// Calcolo errore e controllo
  e_p = sqrt(pow(xDes-x,2)+pow(yDes-y,2));
  if (e_p > 1) // mi muovo solo se l'errore è maggiore di una certa quantità
  {
    tempo = (millis()-t0)/1000; // tempo missione
   
    // assegno velocità secondo legge proporzionale
    v1 = kv1*((xDes-x)*cos(theta) + (yDes-y)*sin(theta)); 

    // Calcolo l'angolo verso il target: scelgo il multiplo di 2PI 
    // più vicino all'orientamento corrente del robot
    thetaDes = atan2(yDes-y,xDes-x) + nGiri*2*PI;
    if (abs(thetaDes+2*PI-theta) < abs(thetaDes-theta))
    {
      thetaDes = thetaDes+2*PI;
      nGiri += 1;
    }
    else
    {
      if (abs(thetaDes-2*PI-theta) < abs(thetaDes-theta))
      {
        thetaDes = thetaDes-2*PI;
        nGiri += -1;
      }
    }

    // assegno velocità angolare secondo legge proporzionale    
    v2 = kv2*(thetaDes-theta);
  }
  else // se sono già abbastanza vicino al target non mi muovo
  {
    v1 = 0;
    v2 = 0;
  }

// Calcolo velocità angolare ruote e loro accelerazioni
  omegaRp = ((v1+v2*d/2)/r-omegaR)/dt;
  omegaLp = ((v1-v2*d/2)/r-omegaL)/dt;
  omegaR = (v1+v2*d/2)/r;
  omegaL = (v1-v2*d/2)/r;

// Per analizzare robustezza legge proporzionale modificare i coefficienti unitari: si introducono perturbazioni dovute a incertezze parametriche
  v1 = (1*r*omegaR+1*r*omegaL)/2;
  v2 = (1*r*omegaR-1*r*omegaL)/(1*d);
  
// Equazione cinematica uniciclo  
  x = x + v1*cos(theta)*dt;
  y = y + v1*sin(theta)*dt;
  theta = theta + v2*dt;
  
  // disegno uniciclo  
  robot(x,y,theta); 
  
  popMatrix();

  textSize(20);
  fill(0,0,255);
  text("v1 (pixel/s) = ",10,20); 
  text(v1,200,20);
  text("v2 (gradi/s) = ",10,50); 
  text(v2*180/PI,200,50);
  
  fill(255,0,0);  
  text("x = ",10,160); 
  text(x,80,160);
  text("y = ",10,190); 
  text(y,80,190);
  text("theta = ",10,220); 
  text(atan2(sin(theta),cos(theta))*180/PI,100,220);  

  fill(255,255,0);
  text("tempo = ",10,110); 
  text(tempo,120,110);  

  fill(0,0,255);
  text("omegaR (gradi/s) = ",650,20); 
  text(omegaL*180/PI,900,20);
  text("omegaL (gradi/s) = ",650,50); 
  text(omegaL*180/PI,900,50);

  fill(255);
  text("nGiri = ",1050,20); 
  text(nGiri,1150,20);  
  
  fill(256,0,0);
  text("xDes = ",700,100); 
  text(xDes,800,100);
  text("yDes = ",700,130); 
  text(yDes,800,130);

  fill(0,255,0);
  text("wRpunto = ",320,20); 
  text(omegaRp,450,20);
  text("wLpunto = ",320,50); 
  text(omegaLp,450,50);

}

void robot(float x, float y, float theta)
{
// funzione che disegna uniciclo in (x,y) con orientamento theta    
  pushMatrix();
  translate(x,-y); // mi porto nella posizione (x,y) dell'uniciclo
  rotate(-theta); // con l'orientamento corretto
  fill(255,0,0);
  ellipse(0,0,2*R,2*R); // il robot
  fill(0,0,255);  
  rect(-r,-d/2-W/2,2*r,W);
  rect(-r,d/2-W/2,2*r,W);
  fill(255);
  ellipse(.8*R,0,.2*R,.2*R);
  ellipse(-.8*R,0,.2*R,.2*R);  
  fill(0,255,0);
  triangle(-.1*R,.3*R,-.1*R,-.3*R,.5*R,0);
  popMatrix();
}