/*
Questo sketch disegna un robot mobile di tipo 'uniciclo'
e permette di muoverlo manualmente nel seguente modo:
freccia SU = accelerazione in avanti;
freccia GIU = accelerazione all'indietro;
freccia SINISTRA = accelerazione verso sinistra;
freccia DESTRA = accelerazione verso destra;
tasto '+'  -> aumento coefficiente accelerazione DeltaV
tasto '-'  -> diminuisco coefficiente  accelerazione DeltaV
tasto 's'  -> stop uniciclo

Sulla schermata vengono riportati: le coordinate
(x,y,theta) dell'uniciclo, le velocità angolari 
(omegaR, omegaL) delle due ruote e le loro
accelerazioni (omegaRp,omegaLp). Inoltre si riportano 
la velocità longitudinale v1, quella angolare v2
e il coefficiente di accelerazione DeltaV.
*/

// Coordinate attuali uniciclo
float x = 0;
float y = 0;
float theta = 0;

// Caratteristiche fisiche uniciclo
float r = 8; // raggio ruote attuate in pixel
float d = 25; // distanza tra le ruote attuate in pixel
float W = 5; // spessore ruote attuate in pixel
float R = 1.2*sqrt(pow(r,2)+pow(d/2+W/2,2)); // raggio robot

float dt = (float) 1/60; // tempo di campionamento

float v1 = 0; // velocità lineare robot
float v2 = 0; // velocità rotazionale robot

float omegaR = 0; // velocità angolare ruota destra
float omegaL = 0; // velocità angolare ruota sinistra
float omegaRp,omegaLp; // accelerazioni angolari delle ruote

float DeltaV = 1; // coefficiente cambiamento velocità

void setup() 
{
  size(1250,700);
}

void draw() 
{
  
  background(0);

  pushMatrix();
  translate(width/2,height/2); // metto l'origine al centro della finestra
  
  if (keyPressed)
  {
    if (keyCode == UP) // aumento velocità longitudinale
    {
      v1 += 5*DeltaV;
    }
    if (keyCode == DOWN)  // diminuisco velocità longitudinale
    {
      v1 -= 5*DeltaV;
    }
    if (keyCode == LEFT)  // aumento velocità angolare
    {
      v2 += .2*DeltaV;
    }
    if (keyCode == RIGHT)  // diminuisco velocità angolare
    {
      v2 -= .2*DeltaV;
    }
    if (key == '+') // aumento coefficiente accelerazione
    {
      DeltaV += .01;
    }
    if (key == '-') // diminuisco coefficiente accelerazione
    {
      DeltaV = max(.01,DeltaV-.01);
    }
    if ((key == 's') || (key == 'S'))
    {
      v1 = 0;
      v2 = 0;
    }
  }    

// Cinematica uniciclo
  x = x + v1*cos(theta)*dt;
  y = y + v1*sin(theta)*dt;
  theta = theta + v2*dt;
  
// disegno uniciclo  
  robot(x,y,theta);

  popMatrix();

// Calcolo velocità angolare ruote e loro accelerazioni
  omegaRp = ((v1+v2*d/2)/r-omegaR)/dt;
  omegaLp = ((v1-v2*d/2)/r-omegaL)/dt;  
  omegaR = (v1+v2*d/2)/r;
  omegaL = (v1-v2*d/2)/r;
  
  textSize(25);
  fill(255);
  text("v1 (pixel/s) = ",10,20); 
  text(v1,200,20);
  text("v2 (gradi/s) = ",10,50); 
  text(v2*180/PI,200,50);
  
  text("x = ",10,160); 
  text(x,80,160);
  text("y = ",10,190); 
  text(y,80,190);
  text("theta = ",10,220); 
  text(atan2(sin(theta),cos(theta))*180/PI,100,220);  

  text("omegaR (gradi/s) = ",650,20); 
  text(omegaR*180/PI,900,20);
  text("omegaL (gradi/s) = ",650,50); 
  text(omegaL*180/PI,900,50);

  text("omegaRpunto (rad/s^2) = ",650,100); 
  text(omegaRp,1000,100);
  text("omegaLpunto (rad/s^2) = ",650,130); 
  text(omegaLp,1000,130);

  text("DeltaV = ",360,20); 
  text(DeltaV,470,20);
  

}

void robot(float x, float y, float theta)
{
// funzione che disegna uniciclo in (x,y) con orientamento theta  
  pushMatrix();
  translate(x,-y); // mi porto nella posizione (x,y) dell'uniciclo
  rotate(-theta); // con l'orientamento corretto
  fill(255,0,0);
  ellipse(0,0,2*R,2*R); // il robot
  fill(0,0,255);  
  rect(-r,-d/2-W/2,2*r,W); // ruota attuata sinistra
  rect(-r,d/2-W/2,2*r,W); // ruota attuata destra
  fill(255);
  ellipse(.8*R,0,.2*R,.2*R); // ruota passiva anteriore
  ellipse(-.8*R,0,.2*R,.2*R);  // ruota passiva posteriore
  fill(0,255,0);
  triangle(-.1*R,.3*R,-.1*R,-.3*R,.5*R,0); // triangolo direzionale
  popMatrix();
}
